Index: kernel/forth.c
===================================================================
--- kernel/forth.c	(revision 23)
+++ kernel/forth.c	(working copy)
@@ -796,9 +796,11 @@
 
 static void emit(void)
 {
+#ifndef FCOMPILER
 	cell tmp = POP();
-#ifndef FCOMPILER
 	putchar(tmp);
+#else
+        (void) POP();
 #endif
 }
 
@@ -834,8 +836,13 @@
 
 static void iocfetch(void)
 {
+#ifndef FCOMPILER
 	cell reg = POP();
 	PUSH(inb(reg));
+#else
+        (void)POP();
+        PUSH(0);
+#endif
 }
 
 
@@ -845,8 +852,13 @@
 
 static void iowfetch(void)
 {
+#ifndef FCOMPILER
 	cell reg = POP();
 	PUSH(inw(reg));
+#else
+        (void)POP();
+        PUSH(0);
+#endif
 }
 
 /*
@@ -855,8 +867,13 @@
 
 static void iolfetch(void)
 {
+#ifndef FCOMPILER
 	cell reg = POP();
 	PUSH(inl(reg));
+#else
+        (void)POP();
+        PUSH(0);
+#endif
 }
 
 
@@ -866,10 +883,15 @@
 
 static void iocstore(void)
 {
+#ifndef FCOMPILER
 	cell reg = POP();
 	cell val = POP();
 
 	outb(reg, val);
+#else
+        (void)POP();
+        (void)POP();
+#endif
 }
 
 
@@ -879,10 +901,15 @@
 
 static void iowstore(void)
 {
+#ifndef FCOMPILER
 	cell reg = POP();
 	cell val = POP();
 
 	outw(reg, val);
+#else
+        (void)POP();
+        (void)POP();
+#endif
 }
 
 
@@ -892,10 +919,15 @@
 
 static void iolstore(void)
 {
+#ifndef FCOMPILER
 	ucell reg = POP();
 	ucell val = POP();
 
 	outl(reg, val);
+#else
+        (void)POP();
+        (void)POP();
+#endif
 }
 
 /*
Index: include/libc/string.h
===================================================================
--- include/libc/string.h	(revision 23)
+++ include/libc/string.h	(working copy)
@@ -42,7 +42,7 @@
 extern void 	*memcpy(void * dest,const void *src,size_t count);
 extern void 	*memmove(void * dest,const void *src,size_t count);
 extern int	memcmp(const void * cs,const void * ct,size_t count);
-extern void 	*memscan(void * addr, int c, size_t size);void * memscan(void * addr, int c, size_t size);
+extern void 	*memscan(void * addr, int c, size_t size);
 extern char 	*strstr(const char * s1,const char * s2);
 extern void	*memchr(const void *s, int c, size_t n);
 
Index: include/sparc32/io.h
===================================================================
--- include/sparc32/io.h	(revision 23)
+++ include/sparc32/io.h	(working copy)
@@ -4,7 +4,8 @@
 #include "asm/types.h"
 
 extern unsigned int va_shift; // Set in entry.S
-extern char _start, _data, _end, _vmem, _evmem, _iomem; // Defined in ldscript
+extern char _start, _data, _heap, _eheap, _stack, _estack, _end,
+    _vmem, _evmem,_iomem; // Defined in ldscript
 
 static inline unsigned long
 va2pa(unsigned long va)
@@ -32,11 +33,12 @@
 
 struct mem;
 
+void mem_init(struct mem *t, char *begin, char *limit);
 void *mem_alloc(struct mem *t, int size, int align);
 void *mem_zalloc(struct mem *t, int size, int align);
 int map_page(unsigned long va, unsigned long epa, int type);
 void *map_io(unsigned pa, int size);
-void init_mmu_swift();
+void init_mmu_swift(unsigned long base);
 void *dvma_alloc(int size, unsigned int *pphys);
 
 #ifndef BOOTSTRAP
Index: include/openbios/bindings.h
===================================================================
--- include/openbios/bindings.h	(revision 23)
+++ include/openbios/bindings.h	(working copy)
@@ -26,7 +26,6 @@
 
 /* initialization */
 extern int		initialize_forth( void );
-extern void		exception( cell errcode );
 
 /* panic */
 extern int		forth_segv_handler( char *segv_addr );
Index: include/openbios/drivers.h
===================================================================
--- include/openbios/drivers.h	(revision 23)
+++ include/openbios/drivers.h	(working copy)
@@ -21,7 +21,7 @@
 int ob_ide_init(void);
 #endif
 #ifdef CONFIG_DRIVER_ESP
-int ob_esp_init(void);
+int ob_esp_init(unsigned long base);
 #endif
 #ifdef CONFIG_DRIVER_OBIO
 int ob_obio_init(unsigned long slavio_base);
Index: toke/stream.c
===================================================================
--- toke/stream.c	(revision 23)
+++ toke/stream.c	(working copy)
@@ -104,7 +104,7 @@
 		oname = (u8 *)strdup( out_name );
 	else {
 		ext=strrchr(in_name, '.');
-		len=ext ? (ext-in_name) : (unsigned int)strlen(in_name) ;
+		len=ext ? (unsigned int)(ext-in_name) : (unsigned int)strlen(in_name) ;
 		oname=malloc(len+4);
 		memcpy(oname, in_name, len);
 		oname[len] = 0;
Index: config/examples/sparc32_config.xml
===================================================================
--- config/examples/sparc32_config.xml	(revision 23)
+++ config/examples/sparc32_config.xml	(working copy)
@@ -46,21 +46,22 @@
   <option name="CONFIG_SUN_PARTS" type="boolean" value="true"/>
   <option name="CONFIG_FS" type="boolean" value="true"/>
   <option name="CONFIG_HFS" type="boolean" value="false"/>
-  <option name="CONFIG_HFSP" type="boolean" value="false"/>
+  <option name="CONFIG_HFSP" type="boolean" value="true"/>
   <option name="CONFIG_GRUBFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_EXT2FS" type="boolean" value="true"/>
-  <option name="CONFIG_FSYS_FAT" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_JFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_MINIX" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_REISERFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_XFS" type="boolean" value="false"/>
+  <option name="CONFIG_FSYS_FAT" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_JFS" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_MINIX" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_REISERFS" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_XFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_UFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_ISO9660" type="boolean" value="true"/>
-  <option name="CONFIG_FSYS_FFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_VSTAFS" type="boolean" value="false"/>
+  <option name="CONFIG_FSYS_FFS" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_VSTAFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_NTFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_AFFS" type="boolean" value="false"/>
+  <option name="CONFIG_FSYS_AFFS" type="boolean" value="true"/>
   <option name="CONFIG_DEBUG_FS" type="boolean" value="false"/>
+  <option name="CONFIG_DEBUG_EXT2FS" type="boolean" value="false"/>
   
   <!-- Miscellaneous -->
   <option name="CONFIG_LINUXBIOS" type="boolean" value="false"/>
@@ -69,7 +70,7 @@
   <option name="CONFIG_DRIVER_SBUS" type="boolean" value="true"/>
   <option name="CONFIG_DRIVER_OBIO" type="boolean" value="true"/>
   <option name="CONFIG_DRIVER_ESP" type="boolean" value="true"/>
-  <option name="CONFIG_DRIVER_FLOPPY" type="boolean" value="false"/>
+  <option name="CONFIG_DRIVER_FLOPPY" type="boolean" value="true"/>
 
 </config>
 
Index: config/examples/sparc32_rules.xml
===================================================================
--- config/examples/sparc32_rules.xml	(revision 23)
+++ config/examples/sparc32_rules.xml	(working copy)
@@ -27,6 +27,8 @@
 endif
 
 CFLAGS := -Os -Wall -W -DNATIVE_BITWIDTH_EQUALS_HOST_BITWIDTH -USWAP_ENDIANNESS -fno-builtin -g
+CFLAGS+= -Wredundant-decls -Wshadow -Wpointer-arith -Wstrict-prototypes -Wmissing-declarations
+CFLAGS+= -Wundef -Wendif-labels -Wstrict-aliasing
 INCLUDES := -Iinclude -Ikernel/include -I$(ODIR)/target/include
 
 #
Index: config/examples/cross-sparc32_config.xml
===================================================================
--- config/examples/cross-sparc32_config.xml	(revision 23)
+++ config/examples/cross-sparc32_config.xml	(working copy)
@@ -46,21 +46,22 @@
   <option name="CONFIG_SUN_PARTS" type="boolean" value="true"/>
   <option name="CONFIG_FS" type="boolean" value="true"/>
   <option name="CONFIG_HFS" type="boolean" value="false"/>
-  <option name="CONFIG_HFSP" type="boolean" value="false"/>
+  <option name="CONFIG_HFSP" type="boolean" value="true"/>
   <option name="CONFIG_GRUBFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_EXT2FS" type="boolean" value="true"/>
-  <option name="CONFIG_FSYS_FAT" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_JFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_MINIX" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_REISERFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_XFS" type="boolean" value="false"/>
+  <option name="CONFIG_FSYS_FAT" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_JFS" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_MINIX" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_REISERFS" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_XFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_UFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_ISO9660" type="boolean" value="true"/>
-  <option name="CONFIG_FSYS_FFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_VSTAFS" type="boolean" value="false"/>
+  <option name="CONFIG_FSYS_FFS" type="boolean" value="true"/>
+  <option name="CONFIG_FSYS_VSTAFS" type="boolean" value="true"/>
   <option name="CONFIG_FSYS_NTFS" type="boolean" value="false"/>
-  <option name="CONFIG_FSYS_AFFS" type="boolean" value="false"/>
+  <option name="CONFIG_FSYS_AFFS" type="boolean" value="true"/>
   <option name="CONFIG_DEBUG_FS" type="boolean" value="false"/>
+  <option name="CONFIG_DEBUG_EXT2FS" type="boolean" value="false"/>
   
   <!-- Miscellaneous -->
   <option name="CONFIG_LINUXBIOS" type="boolean" value="false"/>
@@ -69,7 +70,7 @@
   <option name="CONFIG_DRIVER_SBUS" type="boolean" value="true"/>
   <option name="CONFIG_DRIVER_OBIO" type="boolean" value="true"/>
   <option name="CONFIG_DRIVER_ESP" type="boolean" value="true"/>
-  <option name="CONFIG_DRIVER_FLOPPY" type="boolean" value="false"/>
+  <option name="CONFIG_DRIVER_FLOPPY" type="boolean" value="true"/>
 
 </config>
 
Index: config/examples/cross-sparc32_rules.xml
===================================================================
--- config/examples/cross-sparc32_rules.xml	(revision 23)
+++ config/examples/cross-sparc32_rules.xml	(working copy)
@@ -21,6 +21,8 @@
 STRIP := sparc-linux-strip
 RANLIB := sparc-linux-ranlib
 CFLAGS := -Os -Wall -W -DNATIVE_BITWIDTH_EQUALS_HOST_BITWIDTH -USWAP_ENDIANNESS -fno-builtin -g
+CFLAGS+= -Wredundant-decls -Wshadow -Wpointer-arith -Wstrict-prototypes -Wmissing-declarations
+CFLAGS+= -Wundef -Wendif-labels -Wstrict-aliasing
 INCLUDES := -Iinclude -Ikernel/include -I$(ODIR)/target/include
 
 #
Index: modules/disk-label.c
===================================================================
--- modules/disk-label.c	(revision 23)
+++ modules/disk-label.c	(working copy)
@@ -63,7 +63,9 @@
 	s = parstr;
 	filename = NULL;
 	if( s ) {
-		if( *s == '-' || isdigit(*s) ) {
+            if( *s == '-' || isdigit(*s) ||
+                (*s >= 'a' && *s < ('a' + 8)
+                 && (*(s + 1) == ',' || *(s + 1) == '\0'))) {
 			if( (s=strpbrk(parstr,",")) ) {
 				filename = s+1;
 				*s = 0;
Index: modules/nvram.c
===================================================================
--- modules/nvram.c	(revision 23)
+++ modules/nvram.c	(working copy)
@@ -49,7 +49,7 @@
 static unsigned int
 nvpart_checksum( nvpart_t* hdr )
 {
-	unsigned char *p = (char*)hdr;
+	unsigned char *p = (unsigned char*)hdr;
 	int i, val = p[0];
 
 	for( i=2; i<16; i++ ) {
Index: modules/sun-parts.c
===================================================================
--- modules/sun-parts.c	(revision 23)
+++ modules/sun-parts.c	(working copy)
@@ -27,6 +27,7 @@
 typedef struct {
 	ullong		offs;
 	ullong		size;
+	int		type;
 } sunparts_info_t;
 
 DECLARE_NODE( sunparts, INSTALL_OPEN, sizeof(sunparts_info_t), "+/packages/sun-parts" );
@@ -93,15 +94,19 @@
 	int parnum = -1;
 	unsigned char buf[512];
         struct sun_disklabel *p;
-        unsigned int i;
+        unsigned int i, bs;
 
 	DPRINTF("sunparts_open '%s'\n", str );
 
 	if( str ) {
-		parnum = atol(str);
-		if( !strlen(str) )
-			parnum = -1;
-		free( str );
+            if( !strlen(str) )
+                parnum = -1;
+            else if (str[0] >= 'a' && str[0] < ('a' + 8))
+                parnum = str[0] - 'a';
+            else
+                parnum = atol(str);
+       
+            free( str );
 	}
 
 	SEEK( 0 );
@@ -110,10 +115,11 @@
 
 	/* Check Magic */
 	if (!has_sun_part_magic(buf)) {
-		DPRINTF("sun partition magic not found.\n");
+		DPRINTF("Sun partition magic not found.\n");
 		RET(0);
 	}
-	
+
+	bs = 512; // XXX get real block size?
 	/* get partition data */
 	p = (struct sun_disklabel *)buf;
 
@@ -125,10 +131,16 @@
                     parnum = i;
             }
         }
+
+        if (parnum < 0)
+            parnum = 0;
+
+	DPRINTF("Selected partition %d\n", parnum);
         di->offs = (llong)__be32_to_cpu(p->partitions[parnum].start_cylinder) *
-            __be16_to_cpu(p->ntrks) * __be16_to_cpu(p->nsect);
+            __be16_to_cpu(p->ntrks) * __be16_to_cpu(p->nsect) * bs;
 
-        di->size = (llong)__be32_to_cpu(p->partitions[parnum].num_sectors) ;
+        di->size = (llong)__be32_to_cpu(p->partitions[parnum].num_sectors) * bs;
+        di->type = __be32_to_cpu(p->infos[parnum].id);
         DPRINTF("Found Sun partition table, offs %d size %d\n", (int)di->offs, (int)di->size);
 
         RET( -1 );
@@ -147,7 +159,8 @@
 static void
 sunparts_get_info( sunparts_info_t *di )
 {
-	PUSH( -1 );		/* no type */
+	DPRINTF("Sun get_info\n");
+	PUSH( di->type );
 	DPUSH( di->offs );
 	DPUSH( di->size );
 }
Index: fs/grubfs/fsys_ext2fs.c
===================================================================
--- fs/grubfs/fsys_ext2fs.c	(revision 23)
+++ fs/grubfs/fsys_ext2fs.c	(working copy)
@@ -19,10 +19,19 @@
 
 #ifdef FSYS_EXT2FS
 
+#include "openbios/config.h"
 #include "shared.h"
 #include "filesys.h"
 #include <libc/byteorder.h>
 
+#ifdef CONFIG_DEBUG_EXT2FS
+#define DPRINTF(fmt, args...)                   \
+    do { printk(fmt , ##args); } while (0)
+#define printf printk
+#else
+#define DPRINTF(fmt, args...)
+#endif
+
 static int mapblock1, mapblock2;
 
 /* sizes are always in bytes, BLOCK values are always in DEV_BSIZE (sectors) */
@@ -233,7 +242,7 @@
 #define S_ISREG(m)      (((m) & S_IFMT) == S_IFREG)
 #define S_ISDIR(m)      (((m) & S_IFMT) == S_IFDIR)
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
 void
 dump_super(struct ext2_super_block *s)
 {
@@ -316,10 +325,14 @@
        && (! IS_PC_SLICE_TYPE_BSD_WITH_FS (current_slice, FS_OTHER)))
       || part_length < (SBLOCK + (sizeof (struct ext2_super_block) / DEV_BSIZE))
       || !devread (SBLOCK, 0, sizeof (struct ext2_super_block),
-		   (char *) SUPERBLOCK)
-      || __le16_to_cpu(SUPERBLOCK->s_magic) != EXT2_SUPER_MAGIC)
+		   (char *) SUPERBLOCK))
       retval = 0;
 
+  DPRINTF("ext2fs: magic 0x%x\n", __le16_to_cpu(SUPERBLOCK->s_magic));
+
+  if (__le16_to_cpu(SUPERBLOCK->s_magic) != EXT2_SUPER_MAGIC)
+      retval = 0;
+
   return retval;
 }
 
@@ -327,11 +340,11 @@
 static int
 ext2_rdfsb (int fsblock, char * buffer)
 {
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
   printf ("ext2_rdfsb: fsblock %d, devblock %d, size %d\n", fsblock, 
 	  fsblock * (EXT2_BLOCK_SIZE (SUPERBLOCK) / DEV_BSIZE), 
 	  EXT2_BLOCK_SIZE (SUPERBLOCK));
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
   return devread (fsblock * (EXT2_BLOCK_SIZE (SUPERBLOCK) / DEV_BSIZE), 0,
 		  EXT2_BLOCK_SIZE (SUPERBLOCK), (char *) buffer);
 }
@@ -345,16 +358,16 @@
 ext2fs_block_map (int logical_block)
 {
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
   printf ("ext2fs_block_map(%d)\n", logical_block);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 
   /* if it is directly pointed to by the inode, return that physical addr */
   if (logical_block < EXT2_NDIR_BLOCKS)
     {
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       printf ("ext2fs_block_map: returning %d\n", __le32_to_cpu(INODE->i_block[logical_block]));
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
       return __le32_to_cpu(INODE->i_block[logical_block]);
     }
   /* else */
@@ -436,20 +449,20 @@
   int ret = 0;
   int size = 0;
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
   printf("ext2fs_read(0x%x, %d)\n", buf, len);
   dump_inode(INODE);
   dump_inode_data((unsigned char *)INODE, sizeof (struct ext2_inode));
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
   while (len > 0)
     {
       /* find the (logical) block component of our location */
       logical_block = filepos >> EXT2_BLOCK_SIZE_BITS (SUPERBLOCK);
       offset = filepos & (EXT2_BLOCK_SIZE (SUPERBLOCK) - 1);
       map = ext2fs_block_map (logical_block);
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       printf ("map=%d\n", map);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
       if (map < 0)
 	break;
 
@@ -539,27 +552,27 @@
      the directory known pointed to by current_ino (if any)
    */
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
   printf("****** ext2fs_dir(%s)\n", dirname);
   dump_super(SUPERBLOCK);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 
   while (1)
     {
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       printf ("ext2fs_dir: inode %d\n", current_ino);
       printf ("ext2fs_dir: dirname=%s\n", dirname);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 
       /* look up an inode */
       group_id = (current_ino - 1) / __le32_to_cpu(SUPERBLOCK->s_inodes_per_group);
       group_desc = group_id >> log2 (EXT2_DESC_PER_BLOCK (SUPERBLOCK));
       desc = group_id & (EXT2_DESC_PER_BLOCK (SUPERBLOCK) - 1);
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       printf ("ext2fs_dir: ipg=%d, dpb=%d\n", __le32_to_cpu(SUPERBLOCK->s_inodes_per_group),
 	      EXT2_DESC_PER_BLOCK (SUPERBLOCK));
       printf ("ext2fs_dir: group_id=%d group_desc=%d desc=%d\n", group_id, group_desc, desc);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
       if (!ext2_rdfsb (
 			(WHICH_SUPER + group_desc + __le32_to_cpu(SUPERBLOCK->s_first_data_block)),
 			(char*) GROUP_DESC))
@@ -567,21 +580,21 @@
 	  return 0;
 	}
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       dump_group_desc(GROUP_DESC);
-#endif /* E2DEBUG */	
+#endif /* CONFIG_DEBUG_EXT2FS */	
 
       gdp = GROUP_DESC;
       ino_blk = __le32_to_cpu(gdp[desc].bg_inode_table) +
 	(((current_ino - 1) % __le32_to_cpu(SUPERBLOCK->s_inodes_per_group))
 	 >> log2 (EXT2_BLOCK_SIZE (SUPERBLOCK) / sizeof (struct ext2_inode)));
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       printf ("ext2fs_dir: itab_blk=%d, i_in_grp=%d, log2=%d\n", 
 	 __le32_to_cpu(gdp[desc].bg_inode_table),
 	 ((current_ino - 1) % __le32_to_cpu(SUPERBLOCK->s_inodes_per_group)),
 	 log2 (EXT2_BLOCK_SIZE (SUPERBLOCK) / sizeof (struct ext2_inode)));
       printf ("ext2fs_dir: inode table fsblock=%d\n", ino_blk);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
       if (!ext2_rdfsb (ino_blk, (int) INODE))
 	{
 	  return 0;
@@ -593,7 +606,7 @@
       raw_inode = INODE +
 	((current_ino - 1)
 	 & (EXT2_BLOCK_SIZE (SUPERBLOCK) / sizeof (struct ext2_inode) - 1));
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       printf ("ext2fs_dir: ipb=%d, sizeof(inode)=%d\n",
 	      (EXT2_BLOCK_SIZE (SUPERBLOCK) / sizeof (struct ext2_inode)),
 	      sizeof (struct ext2_inode));
@@ -602,15 +615,15 @@
       dump_inode(raw_inode);
       dump_inode_data((unsigned char *)INODE, EXT2_BLOCK_SIZE(SUPERBLOCK));
       printf ("ext2fs_dir: first word=%x\n", *((int *) raw_inode));
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 
       /* copy inode to fixed location */
       memmove ((void *) INODE, (void *) raw_inode, sizeof (struct ext2_inode));
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
       dump_inode(INODE);
       printf ("ext2fs_dir: first word=%x\n", *((int *) INODE));
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 
       /* If we've got a symbolic link, then chase it. */
       if (S_ISLNK (__le16_to_cpu(INODE->i_mode)))
@@ -659,7 +672,7 @@
 	      memmove (linkbuf, (char *) INODE->i_block, len);
 	    }
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
 	  printf ("ext2fs_dir: symlink=%s\n", linkbuf);
 #endif
 
@@ -719,9 +732,9 @@
       do
 	{
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
 	  printf ("ext2fs_dir: dirname=%s, rest=%s, loc=%d\n", dirname, rest, loc);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 
 	  /* if our location/byte offset into the directory exceeds the size,
 	     give up */
@@ -748,9 +761,9 @@
 	     for, now we have to translate that to the physical (fs) block on
 	     the disk */
 	  map = ext2fs_block_map (blk);
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
 	  printf ("ext2fs_dir: fs block=%d\n", map);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 	  mapblock2 = -1;
 	  if ((map < 0) || !ext2_rdfsb (map, DATABLOCK2))
 	    {
@@ -765,11 +778,11 @@
 
 	  /* NOTE: ext2fs filenames are NOT null-terminated */
 
-#ifdef E2DEBUG
+#ifdef CONFIG_DEBUG_EXT2FS
 	  printf ("ext2fs_dir: directory entry ino=%d\n", __le32_to_cpu(dp->inode));
 	  if (__le32_to_cpu(dp->inode))
 	    printf ("entry=%s\n", dp->name);
-#endif /* E2DEBUG */
+#endif /* CONFIG_DEBUG_EXT2FS */
 
 	  if (__le32_to_cpu(dp->inode))
 	    {
Index: fs/grubfs/grubfs_fs.c
===================================================================
--- fs/grubfs/grubfs_fs.c	(revision 23)
+++ fs/grubfs/grubfs_fs.c	(working copy)
@@ -303,7 +303,7 @@
 {
 	if (filepos < 0 || filepos > filemax)
 		filepos = filemax;
-	if (len < 0 || len > filemax-filepos)
+	if (len > filemax-filepos)
 		len = filemax - filepos;
 	errnum = 0;
 	return curfs->fsys->read_func( buf, len );
Index: fs/grubfs/iso9660.h
===================================================================
--- fs/grubfs/iso9660.h	(revision 23)
+++ fs/grubfs/iso9660.h	(working copy)
@@ -158,10 +158,13 @@
 
 #define	RRMAGIC(c1, c2)	((c1)|(c2) << 8)
 
-#define	CHECK2(ptr, c1, c2) \
-	(*(unsigned short *)(ptr) == (((c1) | (c2) << 8) & 0xFFFF))
-#define	CHECK4(ptr, c1, c2, c3, c4) \
-	(*(unsigned long *)(ptr) == ((c1) | (c2)<<8 | (c3)<<16 | (c4)<<24))
+#define	CHECK2(ptr, c1, c2)                     \
+    (*(ptr) == (c1) && *(ptr + 1) == (c2))
+#define	CHECK4(ptr, c1, c2, c3, c4)                                     \
+    (*(unsigned char *)(ptr) == (c1) &&                                 \
+     *((unsigned char *)(ptr) + 1) == (c2) &&                           \
+     *((unsigned char *)(ptr) + 2) == (c3) &&                           \
+     *((unsigned char *)(ptr) + 3) == (c4))
 
 #endif /* !ASM_FILE */
 
Index: libc/diskio.c
===================================================================
--- libc/diskio.c	(revision 23)
+++ libc/diskio.c	(working copy)
@@ -184,7 +184,8 @@
 
 	DPUSH( offs );
 	call_package( fdp->seek_xt, fdp->ih );
-	return ((POP() >= 0)? 0 : -1);
+        (void) POP();
+	return 0;
 }
 
 llong
Index: arch/sparc32/libgcc/__lshrdi3.c
===================================================================
--- arch/sparc32/libgcc/__lshrdi3.c	(revision 0)
+++ arch/sparc32/libgcc/__lshrdi3.c	(revision 0)
@@ -0,0 +1,62 @@
+/* lshrdi3.c extracted from gcc-2.7.2/libgcc2.c which is: */
+/* Copyright (C) 1989, 1992, 1993, 1994, 1995 Free Software Foundation, Inc.
+
+This file is part of GNU CC.
+
+GNU CC is free software; you can redistribute it and/or modify
+it under the terms of the GNU General Public License as published by
+the Free Software Foundation; either version 2, or (at your option)
+any later version.
+
+GNU CC is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+GNU General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with GNU CC; see the file COPYING.  If not, write to
+the Free Software Foundation, 59 Temple Place - Suite 330,
+Boston, MA 02111-1307, USA.  */
+
+#define BITS_PER_UNIT 8
+
+typedef 	 int SItype	__attribute__ ((mode (SI)));
+typedef unsigned int USItype	__attribute__ ((mode (SI)));
+typedef		 int DItype	__attribute__ ((mode (DI)));
+typedef int word_type __attribute__ ((mode (__word__)));
+
+struct DIstruct {SItype high, low;};
+
+typedef union
+{
+  struct DIstruct s;
+  DItype ll;
+} DIunion;
+
+DItype
+__lshrdi3 (DItype u, word_type b)
+{
+  DIunion w;
+  word_type bm;
+  DIunion uu;
+
+  if (b == 0)
+    return u;
+
+  uu.ll = u;
+
+  bm = (sizeof (SItype) * BITS_PER_UNIT) - b;
+  if (bm <= 0)
+    {
+      w.s.high = 0;
+      w.s.low = (USItype)uu.s.high >> -bm;
+    }
+  else
+    {
+      USItype carries = (USItype)uu.s.high << bm;
+      w.s.high = (USItype)uu.s.high >> b;
+      w.s.low = ((USItype)uu.s.low >> b) | carries;
+    }
+
+  return w.ll;
+}
Index: arch/sparc32/libgcc/build.xml
===================================================================
--- arch/sparc32/libgcc/build.xml	(revision 23)
+++ arch/sparc32/libgcc/build.xml	(working copy)
@@ -7,6 +7,7 @@
   <object source="__udivdi3.c"/>
   <object source="__udivmoddi4.c"/>
   <object source="__umoddi3.c"/>
+  <object source="__lshrdi3.c"/>
  </library>
 
 </build>
Index: arch/sparc32/ldscript
===================================================================
--- arch/sparc32/ldscript	(revision 23)
+++ arch/sparc32/ldscript	(working copy)
@@ -1,6 +1,9 @@
 OUTPUT_FORMAT(elf32-sparc)
 OUTPUT_ARCH(sparc)
 
+/* Qemu ELF loader can't handle very complex files, so we put ELFBoot
+info to rodata and put initctx to data.*/
+
 ENTRY(trap_table)
 
 /* Initial load address
@@ -35,6 +38,7 @@
 	sound_drivers_end = .;
 	*(.rodata)
 	*(.rodata.*)
+        *(.note.ELFBoot)
     }
     .data ALIGN(4096): {
         _data = .;
@@ -68,18 +72,12 @@
 	_estack = .;
     }
 
-    .initctx : {
-	/* Initial contents of stack. This MUST BE just after the stack. */
-	*(.initctx)
-    }
-    /* Putting ELF notes near beginning of file might help bootloaders.
-     * We discard .note sections other than .note.ELFBoot,
-     * because some versions of GCC generates useless ones. */
-    .note : { *(.note.ELFBoot) }
-
     . = ALIGN(4096);
     _end = .;
     _iomem = _end + IOMEM_SIZE;
 
-    /DISCARD/ : { *(.comment.*) *(.note.*) }
+    /* We discard .note sections other than .note.ELFBoot,
+     * because some versions of GCC generates useless ones. */
+
+    /DISCARD/ : { *(.comment*) *(.note.*) }
 }
Index: arch/sparc32/elfload.c
===================================================================
--- arch/sparc32/elfload.c	(revision 23)
+++ arch/sparc32/elfload.c	(working copy)
@@ -76,7 +76,8 @@
 }
 
 static unsigned long process_image_notes(Elf_phdr *phdr, int phnum,
-	unsigned short *sum_ptr)
+                                         unsigned short *sum_ptr,
+                                         unsigned int offset)
 {
     int i;
     char *buf = NULL;
@@ -90,7 +91,7 @@
 	if (phdr[i].p_type != PT_NOTE)
 	    continue;
 	buf = malloc(phdr[i].p_filesz);
-	file_seek(phdr[i].p_offset);
+	file_seek(offset + phdr[i].p_offset);
 	if (lfile_read(buf, phdr[i].p_filesz) != phdr[i].p_filesz) {
 	    printf("Can't read note segment\n");
 	    goto out;
@@ -132,7 +133,8 @@
 }
 
 static int load_segments(Elf_phdr *phdr, int phnum,
-	unsigned long checksum_offset)
+                         unsigned long checksum_offset,
+                         unsigned int offset)
 {
     unsigned long bytes;
     //unsigned int start_time, time;
@@ -145,7 +147,7 @@
 	    continue;
 	debug("segment %d addr:%#x file:%#x mem:%#x ",
               i, addr_fixup(phdr[i].p_paddr), phdr[i].p_filesz, phdr[i].p_memsz);
-	file_seek(phdr[i].p_offset);
+	file_seek(offset + phdr[i].p_offset);
 	debug("loading... ");
 	if (lfile_read(phys_to_virt(addr_fixup(phdr[i].p_paddr)), phdr[i].p_filesz)
 		!= phdr[i].p_filesz) {
@@ -309,16 +311,24 @@
     Elf_Bhdr *boot_notes = NULL;
     int retval = -1;
     int image_retval;
+    unsigned int offset;
 
     image_name = image_version = 0;
 
     if (!file_open(filename))
 	goto out;
 
-    if (lfile_read(&ehdr, sizeof ehdr) != sizeof ehdr) {
-	debug("Can't read ELF header\n");
-	retval = LOADER_NOT_SUPPORT;
-	goto out;
+    for (offset = 0; offset < 16 * 512; offset += 512) {
+        if (lfile_read(&ehdr, sizeof ehdr) != sizeof ehdr) {
+            debug("Can't read ELF header\n");
+            retval = LOADER_NOT_SUPPORT;
+            goto out;
+        }
+
+        if (ehdr.e_ident[EI_MAG0] == ELFMAG0)
+            break;
+
+        file_seek(offset);
     }
 
     if (ehdr.e_ident[EI_MAG0] != ELFMAG0
@@ -339,7 +349,7 @@
 
     phdr_size = ehdr.e_phnum * sizeof *phdr;
     phdr = malloc(phdr_size);
-    file_seek(ehdr.e_phoff);
+    file_seek(offset + ehdr.e_phoff);
     if (lfile_read(phdr, phdr_size) != phdr_size) {
 	printf("Can't read program header\n");
 	goto out;
@@ -348,14 +358,14 @@
     if (!check_mem_ranges(info, phdr, ehdr.e_phnum))
 	goto out;
 
-    checksum_offset = process_image_notes(phdr, ehdr.e_phnum, &checksum);
+    checksum_offset = process_image_notes(phdr, ehdr.e_phnum, &checksum, offset);
 
     printf("Loading %s", image_name ? image_name : "image");
     if (image_version)
 	printf(" version %s", image_version);
     printf("...\n");
 
-    if (!load_segments(phdr, ehdr.e_phnum, checksum_offset))
+    if (!load_segments(phdr, ehdr.e_phnum, checksum_offset, offset))
 	goto out;
     
     if (checksum_offset) {
@@ -373,12 +383,13 @@
 #if 1
     {
         extern unsigned int qemu_mem_size;
+        extern char boot_device;
         void *init_openprom(unsigned long memsize, const char *cmdline, char boot_device);
 
         int (*entry)(const void *romvec, int p2, int p3, int p4, int p5);
         const void *romvec;
 
-        romvec = init_openprom(qemu_mem_size, cmdline, 'c');
+        romvec = init_openprom(qemu_mem_size, cmdline, boot_device);
         entry = (void *) addr_fixup(ehdr.e_entry);
         image_retval = entry(romvec, 0, 0, 0, 0);
     }
Index: arch/sparc32/aoutload.c
===================================================================
--- arch/sparc32/aoutload.c	(revision 23)
+++ arch/sparc32/aoutload.c	(working copy)
@@ -57,16 +57,23 @@
     int image_retval;
     struct exec ehdr;
     unsigned long start, size;
+    unsigned int offset;
 
     image_name = image_version = 0;
 
     if (!file_open(filename))
 	goto out;
 
-    if (lfile_read(&ehdr, sizeof ehdr) != sizeof ehdr) {
-	debug("Can't read a.out header\n");
-	retval = LOADER_NOT_SUPPORT;
-	goto out;
+    for (offset = 0; offset < 16 * 512; offset += 512) {
+        if (lfile_read(&ehdr, sizeof ehdr) != sizeof ehdr) {
+            debug("Can't read a.out header\n");
+            retval = LOADER_NOT_SUPPORT;
+            goto out;
+        }
+        if (!N_BADMAG(ehdr))
+            break;
+
+        file_seek(offset);
     }
 
     if (N_BADMAG(ehdr)) {
@@ -76,9 +83,9 @@
     }
 
     if (N_MAGIC(ehdr) == NMAGIC) {
-        size = N_DATADDR(ehdr) + ehdr.a_data;
+        size = addr_fixup(N_DATADDR(ehdr)) + addr_fixup(ehdr.a_data);
     } else {
-        size = ehdr.a_text + ehdr.a_data;
+        size = addr_fixup(ehdr.a_text) + addr_fixup(ehdr.a_data);
     }
 
     start = 0x4000; // N_TXTADDR(ehdr);
@@ -88,38 +95,39 @@
 
     printf("Loading a.out %s...\n", image_name ? image_name : "image");
 
-    file_seek(N_TXTOFF(ehdr));
+    file_seek(offset + N_TXTOFF(ehdr));
 
     if (N_MAGIC(ehdr) == NMAGIC) {
         if (lfile_read(start, ehdr.a_text) != ehdr.a_text) {
-            printf("Can't read program text segment (size 0x%x)\n", ehdr.a_text);
+            printf("Can't read program text segment (size 0x%lx)\n", ehdr.a_text);
             goto out;
         }
         if (lfile_read(start + N_DATADDR(ehdr), ehdr.a_data) != ehdr.a_data) {
-            printf("Can't read program data segment (size 0x%x)\n", ehdr.a_data);
+            printf("Can't read program data segment (size 0x%lx)\n", ehdr.a_data);
             goto out;
         }
     } else {
         if (lfile_read(start, size) != size) {
-            printf("Can't read program (size 0x%x)\n", size);
+            printf("Can't read program (size 0x%lx)\n", size);
             goto out;
         }
     }
     
     debug("Loaded %lu bytes\n", size);
 
-    debug("entry point is %#x\n", start);
+    debug("entry point is %#lx\n", start);
     printf("Jumping to entry point...\n");
 
 #if 1
     {
         extern unsigned int qemu_mem_size;
+        extern char boot_device;
         void *init_openprom(unsigned long memsize, const char *cmdline, char boot_device);
 
         int (*entry)(const void *romvec, int p2, int p3, int p4, int p5);
         const void *romvec;
 
-        romvec = init_openprom(qemu_mem_size, cmdline, 'c');
+        romvec = init_openprom(qemu_mem_size, cmdline, boot_device);
         entry = (void *) addr_fixup(start);
         image_retval = entry(romvec, 0, 0, 0, 0);
     }
Index: arch/sparc32/openbios.c
===================================================================
--- arch/sparc32/openbios.c	(revision 23)
+++ arch/sparc32/openbios.c	(working copy)
@@ -19,11 +19,14 @@
 void boot(void);
 void ob_ide_init(void);
 
-static char intdict[256 * 1024];
+#define IOMMU_BASE    0x10000000 /* First page of sun4m IOMMU */
+#define SLAVIO_BASE   0x71000000
+#define MACIO_BASE    0x70000000
 
+static unsigned char intdict[256 * 1024];
+
 static void init_memory(void)
 {
-        extern char _heap, _eheap;
 
 	/* push start and end of available memory to the stack
 	 * so that the forth word QUIT can initialize memory 
@@ -36,7 +39,7 @@
 	PUSH((unsigned int)&_eheap);
 }
 
-void exception(cell no)
+void exception(__attribute__((unused))cell no)
 {
 	/* The exception mechanism is used during bootstrap to catch
 	 * build errors. In a running system this is a noop since we
@@ -51,23 +54,28 @@
 
 	modules_init();
 #ifdef CONFIG_DRIVER_SBUS
-        init_mmu_swift();
+        init_mmu_swift(IOMMU_BASE);
    
 	ob_sbus_init();
 #endif
 #ifdef CONFIG_DRIVER_ESP
-	ob_esp_init();
+	ob_esp_init(MACIO_BASE);
 #endif
 #ifdef CONFIG_DRIVER_OBIO
-	ob_obio_init(0x71000000);
+	ob_obio_init(SLAVIO_BASE);
+        nvram_init();
 #endif
 	device_end();
+
 	bind_func("platform-boot", boot );
 }
 
 int openbios(void)
 {
 	extern struct sys_info sys_info;
+        extern struct mem cmem;
+
+        mem_init(&cmem, (char *) &_vmem, (char *)&_evmem);
 #ifdef CONFIG_DEBUG_CONSOLE
 #ifdef CONFIG_DEBUG_CONSOLE_SERIAL
 	uart_init(CONFIG_SERIAL_PORT, CONFIG_SERIAL_SPEED);
Index: arch/sparc32/tree.fs
===================================================================
--- arch/sparc32/tree.fs	(revision 23)
+++ arch/sparc32/tree.fs	(working copy)
@@ -5,16 +5,13 @@
   " SUNW,SparcStation-5" encode-string " name" property
   " SparcStation" encode-string " banner-name" property
   " sun4m" encode-string " compatible" property
-  h# 01800000 encode-int 0 encode-int encode+ 0 encode-int encode+ h# 00000081 encode-int encode+ 
-   0 encode-int encode+ 0 encode-int encode+ 0 encode-int encode+ 0 encode-int encode+
-   " idprom" property \ XXX
+  " /obio/zs@0,100000:a" encode-string " stdin-path" property
+  " /obio/zs@0,100000:a" encode-string " stdout-path" property
   : encode-unit encode-unit-sbus ;
   : decode-unit decode-unit-sbus ;
 
 new-device
   " memory" device-name
-  h# 100000 encode-int " available" property \ XXX
-  h# 0 encode-int h# 10000000 encode-int encode+ h# 00000300 encode-int encode+ " reg" property \ XXX
   external
   : open true ;
   : close ;
@@ -24,8 +21,6 @@
 
 new-device
   " virtual-memory" device-name
-  h# 100000 encode-int " available" property \ XXX
-  h# 0 encode-int h# 10000000 encode-int encode+ h# 00000300 encode-int encode+ " reg" property \ XXX
   external
   : open true ;
   : close ;
@@ -129,7 +124,6 @@
 " /iommu/sbus" find-device
 new-device
   " espdma" device-name
-  h# 4 encode-int h# 08400000 encode-int encode+ h# 00000010 encode-int encode+ " reg" property
   external
   : encode-unit encode-unit-sbus ;
   : decode-unit decode-unit-sbus ;
@@ -178,76 +172,6 @@
   : decode-unit decode-unit-sbus ;
 finish-device
 
-" /obio" find-device
-new-device
-  " SUNW,fdtwo" device-name
-  " block" device-type
-  h# 0 encode-int h# 00400000 encode-int encode+ h# 0000000f encode-int encode+ " reg" property
-  h# 2b encode-int 0 encode-int encode+ " intr" property
-finish-device
-
-" /obio" find-device
-new-device
-  " auxio" device-name
-  h# 0 encode-int h# 00900000 encode-int encode+ h# 00000001 encode-int encode+ " reg" property
-finish-device
-
-" /obio" find-device
-new-device
-  " counter" device-name
-  h# 0 encode-int h# 00d00000 encode-int encode+ h# 00000010 encode-int encode+
-  h# 0 encode-int encode+ h# 00d10000 encode-int encode+ h# 00000010 encode-int encode+
-   " reg" property
-finish-device
-
-" /obio" find-device
-new-device
-  " eeprom" device-name
-  h# 0 encode-int h# 00200000 encode-int encode+ h# 00002000 encode-int encode+ " reg" property
-  " mk48t08" model
-finish-device
-
-" /obio" find-device
-new-device
-  " interrupt" device-name
-  h# 0 encode-int h# 00e00000 encode-int encode+ h# 00000010 encode-int encode+
-  h# 0 encode-int encode+ h# 00e10000 encode-int encode+ h# 00000010 encode-int encode+
-   " reg" property
-finish-device
-
-" /obio" find-device
-new-device
-  " power" device-name
-  h# 0 encode-int h# 00910000 encode-int encode+ h# 00000001 encode-int encode+ " reg" property
-  h# 22 encode-int 0 encode-int encode+ " intr" property
-finish-device
-
-" /obio" find-device
-new-device
-  " slavioconfig" device-name
-  h# 0 encode-int h# 00800000 encode-int encode+ h# 00000001 encode-int encode+ " reg" property
-finish-device
-
-" /obio" find-device
-new-device
-  " zs" device-name
-  " serial" device-type
-  h# 0 encode-int h# 00000000 encode-int encode+ h# 00000008 encode-int encode+ " reg" property
-  1 encode-int " slave" property
-  h# 2c encode-int 0 encode-int encode+ " intr" property
-  -1 encode-int " keyboard" property
-  -1 encode-int " mouse" property
-finish-device
-
-" /obio" find-device
-new-device
-  " zs" device-name
-  " serial" device-type
-  h# 0 encode-int h# 00100000 encode-int encode+ h# 00000008 encode-int encode+ " reg" property
-  0 encode-int " slave" property
-  h# 2c encode-int 0 encode-int encode+ " intr" property
-finish-device
-
 " /options" find-device
   " disk" encode-string " boot-from" property
 
Index: arch/sparc32/sys_info.c
===================================================================
--- arch/sparc32/sys_info.c	(revision 23)
+++ arch/sparc32/sys_info.c	(working copy)
@@ -1,5 +1,6 @@
 #include "openbios/config.h"
 #include "openbios/kernel.h"
+#include "elf_boot.h"
 #include "sys_info.h"
 #include "context.h"
 
@@ -19,6 +20,7 @@
     int i;
     unsigned long long total = 0;
     struct memrange *mmap;
+    extern struct elf_image_note elf_image_notes;
 
     /* Pick up paramters given by bootloader to us */
     //info->boot_type = boot_ctx->eax;
@@ -26,6 +28,8 @@
     info->boot_arg = boot_ctx->param[0];
     //debug("boot eax = %#lx\n", info->boot_type);
     //debug("boot ebx = %#lx\n", info->boot_data);
+    info->boot_type = ELF_BHDR_MAGIC;
+    info->boot_data = virt_to_phys(&elf_image_notes);
     debug("boot arg = %#lx\n", info->boot_arg);
 
     collect_elfboot_info(info);
Index: arch/sparc32/romvec.c
===================================================================
--- arch/sparc32/romvec.c	(revision 23)
+++ arch/sparc32/romvec.c	(working copy)
@@ -61,11 +61,9 @@
 static void (*synch_hook)(void);
 
 static struct linux_romvec romvec0;
-static int prop_mem_reg[3];
-static int prop_mem_avail[3];
-static int prop_vmem_avail[6];
 
-static void doublewalk(unsigned ptab1, unsigned va)
+static void doublewalk(__attribute__((unused)) unsigned int ptab1,
+                       __attribute__((unused)) unsigned int va)
 {
 }
 
@@ -215,14 +213,14 @@
     for (;;) {}
 }
 
-static void obp_abort()
+static void obp_abort(void)
 {
     printk("abort, power off\n");
     outb(0x71910000, 1);
     for (;;) {}
 }
 
-static void obp_halt()
+static void obp_halt(void)
 {
     printk("halt, power off\n");
     outb(0x71910000, 1);
@@ -278,7 +276,7 @@
     unsigned int off;
     unsigned int mva;
 
-    DPRINTF("obp_dumb_mmap: virta 0x%x, which_io %d, paddr 0x%x, sz %d\n", va, which_io, pa, size);
+    DPRINTF("obp_dumb_mmap: virta 0x%x, which_io %d, paddr 0x%x, sz %d\n", (unsigned int)va, which_io, pa, size);
     off = pa & (PAGE_SIZE-1);
     npages = (off + size + (PAGE_SIZE-1)) / PAGE_SIZE;
     pa &= ~(PAGE_SIZE-1);
@@ -296,7 +294,7 @@
 static void obp_dumb_munmap(__attribute__((unused)) char *va,
 			    __attribute__((unused)) unsigned int size)
 {
-    DPRINTF("obp_dumb_munmap: virta 0x%x, sz %d\n", va, size);
+    DPRINTF("obp_dumb_munmap: virta 0x%x, sz %d\n", (unsigned int)va, size);
 }
 
 static int obp_devread(int dev_desc, char *buf, int nbytes)
@@ -360,14 +358,18 @@
     return ret;
 }
 
-static int obp_cpustart(unsigned int whichcpu, int ctxtbl_ptr,
-                        int thiscontext, char *prog_counter)
+static int obp_cpustart(__attribute__((unused))unsigned int whichcpu,
+                        __attribute__((unused))int ctxtbl_ptr,
+                        __attribute__((unused))int thiscontext,
+                        __attribute__((unused))char *prog_counter)
 {
     //int cpu, found;
+#ifdef CONFIG_DEBUG_OBP
     struct linux_prom_registers *smp_ctable = (void *)ctxtbl_ptr;
+#endif
 
     DPRINTF("obp_cpustart: cpu %d, ctxptr 0x%x, ctx %d, pc 0x%x\n", whichcpu,
-            smp_ctable->phys_addr, thiscontext, prog_counter);
+            smp_ctable->phys_addr, thiscontext, (unsigned int)prog_counter);
 #if 0
     found = obp_getprop(whichcpu, "mid", (char *)&cpu);
     if (found == -1)
@@ -413,31 +415,18 @@
     /*
      * Form memory descriptors.
      */
-    totphys[0].theres_more = 0;
+    totphys[0].theres_more = NULL;
     totphys[0].start_adr = (char *) 0;
     totphys[0].num_bytes = memsize;
 
-    totavail[0].theres_more = 0;
-    totavail[0].start_adr = (char*) 0;
+    totavail[0].theres_more = NULL;
+    totavail[0].start_adr = (char *) 0;
     totavail[0].num_bytes = memsize;
 
-    totmap[0].theres_more = 0;
+    totmap[0].theres_more = NULL;
     totmap[0].start_adr = &_start;
     totmap[0].num_bytes = (unsigned long) &_iomem - (unsigned long) &_start;
 
-    prop_mem_reg[0] = 0;
-    prop_mem_reg[1] = 0;
-    prop_mem_reg[2] = memsize;
-    prop_mem_avail[0] = 0;
-    prop_mem_avail[1] = 0;
-    prop_mem_avail[2] = va2pa((unsigned long)&_data) - 1;
-    prop_vmem_avail[0] = 0;
-    prop_vmem_avail[1] = 0;
-    prop_vmem_avail[2] = (unsigned long)&_start - 1;
-    prop_vmem_avail[3] = 0;
-    prop_vmem_avail[4] = 0xffe00000;
-    prop_vmem_avail[5] = 0x00200000;
-
     // Linux wants a R/W romvec table
     romvec0.pv_magic_cookie = LINUX_OPPROM_MAGIC;
     romvec0.pv_romvers = 0;
@@ -473,6 +462,7 @@
     romvec0.pv_v2devops.v2_dev_seek = obp_devseek;
     obp_arg.boot_dev_ctrl = 0;
     obp_arg.boot_dev_unit = 0;
+    obp_arg.dev_partition = 4;
     obp_arg.argv[0] = "sd(0,0,0):d";
 
     switch(boot_device) {
@@ -498,7 +488,7 @@
     }
     obp_arg.argv[1] = cmdline;
     romvec0.pv_v2bootargs.bootpath = &obp_arg.argv[0];
-    romvec0.pv_v2bootargs.bootargs = &cmdline;
+    romvec0.pv_v2bootargs.bootargs = &obp_arg.argv[1];
     romvec0.pv_v2bootargs.fd_stdin = &obp_fd_stdin;
     romvec0.pv_v2bootargs.fd_stdout = &obp_fd_stdout;
 
Index: arch/sparc32/boot.c
===================================================================
--- arch/sparc32/boot.c	(revision 23)
+++ arch/sparc32/boot.c	(working copy)
@@ -16,24 +16,54 @@
 void boot(void);
 
 struct sys_info sys_info;                                                       
+uint32_t kernel_image;
+uint32_t kernel_size;
+uint32_t cmdline;
+uint32_t cmdline_size;
+char boot_device;
 
 void boot(void)
 {
 	char *path=pop_fstr_copy(), *param;
 	
-	// char *param="root=/dev/hda2 console=ttyS0,115200n8 console=tty0";
-	
+        if (kernel_size) {
+            extern unsigned int qemu_mem_size;
+            void *init_openprom(unsigned long memsize, const char *cmdline, char boot_device);
+
+            int (*entry)(const void *romvec, int p2, int p3, int p4, int p5);
+            const void *romvec;
+
+            printk("[sparc] Kernel already loaded\n");
+            romvec = init_openprom(qemu_mem_size, (void *)cmdline, boot_device);
+            entry = (void *) kernel_image;
+            entry(romvec, 0, 0, 0, 0);
+        }
+
 	if(!path) {
+            switch(boot_device) {
+            case 'a':
+                path = "/obio/SUNW,fdtwo";
+                break;
+            case 'c':
+                path = "disk";
+                break;
+            default:
+            case 'd':
                 path = "cdrom";
-		//printk("[sparc] Booting default not supported.\n");
-		//return;
+                break;
+            case 'n':
+                path = "net";
+                break;
+            }
 	}
 
 	param = strchr(path, ' ');
 	if(param) {
 		*param = '\0';
 		param++;
-	}
+	} else if (cmdline_size) {
+            param = (char *)cmdline;
+        }
 	
 	printk("[sparc] Booting file '%s' with parameters '%s'\n",path, param);
 
Index: arch/sparc32/lib.c
===================================================================
--- arch/sparc32/lib.c	(revision 23)
+++ arch/sparc32/lib.c	(working copy)
@@ -32,22 +32,11 @@
 	return i;
 }
 
-// dumb quick memory allocator until we get a decent thing here.
-
-#define MEMSIZE 128*1024
-static char memory[MEMSIZE];
-static void *memptr=memory;
-static int memsize=MEMSIZE;
-
 void *malloc(int size)
 {
-	void *ret=(void *)0;
-	if(memsize>=size) {
-		memsize-=size;
-		ret=memptr;
-		memptr+=size;
-	}
-	return ret;
+    extern struct mem cmem;
+
+    return mem_alloc(&cmem, size, 4);
 }
 
 void free(void *ptr)
Index: arch/sparc32/loadfs.c
===================================================================
--- arch/sparc32/loadfs.c	(revision 23)
+++ arch/sparc32/loadfs.c	(working copy)
@@ -8,26 +8,35 @@
 int file_open(const char *filename)
 {
 	load_fd=open_io(filename);
-	/* if(load_fd!=-1)  */ seek_io(load_fd, 0);
+	if(load_fd >= 0)
+            seek_io(load_fd, 0);
 	return load_fd>-1;
 }
 
 int lfile_read(void *buf, unsigned long len)
 {
-	int ret;
-	ret=read_io(load_fd, buf, len);
+	int ret = 0;
+
+        if (load_fd >= 0)
+            ret=read_io(load_fd, buf, len);
 	return ret;
 }
 
 int file_seek(unsigned long offset)
 {
-	return seek_io(load_fd, offset);
+        if (load_fd >= 0)
+            return seek_io(load_fd, offset);
+        else
+            return -1;
 }
 
 unsigned long file_size(void)
 {
 	llong fpos, fsize;
 	
+        if (load_fd < 0)
+            return 0;
+
 	/* save current position */
 	fpos=tell(load_fd);
 
Index: arch/sparc32/entry.S
===================================================================
--- arch/sparc32/entry.S	(revision 23)
+++ arch/sparc32/entry.S	(working copy)
@@ -64,6 +64,7 @@
         or	%g3, 0x1, %g3
         sta	%g3, [%g2] ASI_M_BYPASS
         add	%g2, 0x4, %g2			! s+0x800
+#if 0
         set     0x40, %g6
         set	((7 << 2) | 2), %g3		! 7 = U: --- S: RWX (main memory)
 1:      sta	%g3, [%g2] ASI_M_BYPASS
@@ -71,6 +72,9 @@
         deccc   %g6
         bne     1b
          nop
+#else
+        add     %g2, 0x100, %g2
+#endif
                                                 ! s+0x900
         add	%g2, 0xa00 - 0x900, %g3		! l3 table for rom at s+0xa00
         add	%g2, 0x0d0, %g2			! s+0x9d0
@@ -182,7 +186,7 @@
 
         /* Zero out our BSS section. */
         set	_bss - 4, %o0           ! First address of BSS
-        set	_end, %o1               ! Last address of BSS
+        set	_estack, %o1               ! Last address of BSS
         ba	2f
          nop
 1:
@@ -192,8 +196,6 @@
         bl	1b
          add	%o0, 0x4, %o0
 
-        set     _estack, %sp
-        mov     0, %fp
         mov     2, %g1
         wr      %g1, 0x0, %wim			! make window 1 invalid
         WRITE_PAUSE
@@ -218,7 +220,7 @@
         wr	%g3, PSR_ET, %psr
         WRITE_PAUSE
 
-        call    openbios
+        call    __switch_context_nosave
          nop
 
         /* We get here when the main context switches back to
Index: arch/sparc32/switch.S
===================================================================
--- arch/sparc32/switch.S	(revision 23)
+++ arch/sparc32/switch.S	(working copy)
@@ -21,62 +21,94 @@
 /* XXX: totally bogus for sparc, need to save and restore all windows */
 __switch_context:
 	/* Save everything in current stack */
+	set	__context, %g1
+        st      %g2, [%g1 + 8]
+        st      %g3, [%g1 + 12]
+        st      %g4, [%g1 + 16]
+        st      %g5, [%g1 + 20]
+        st      %g6, [%g1 + 24]
+        st      %g7, [%g1 + 28]
 
+        st      %o0, [%g1 + 32]
+        st      %o1, [%g1 + 36]
+        st      %o2, [%g1 + 40]
+        st      %o3, [%g1 + 44]
+        st      %o4, [%g1 + 48]
+        st      %o5, [%g1 + 52]
+        st      %o6, [%g1 + 56]
+        st      %o7, [%g1 + 60]
+
+        st      %l0, [%g1 + 64]
+        st      %l1, [%g1 + 68]
+        st      %l2, [%g1 + 72]
+        st      %l3, [%g1 + 76]
+        st      %l4, [%g1 + 80]
+        st      %l5, [%g1 + 84]
+        st      %l6, [%g1 + 88]
+        st      %l7, [%g1 + 92]
+
+        st      %i0, [%g1 + 96]
+        st      %i1, [%g1 + 100]
+        st      %i2, [%g1 + 104]
+        st      %i3, [%g1 + 108]
+        st      %i4, [%g1 + 112]
+        st      %i5, [%g1 + 116]
+        st      %i6, [%g1 + 120]
+        st      %i7, [%g1 + 124]
+
 __switch_context_nosave:
 	/* Interrupts are not allowed... */
 
 	/* Turn on Supervisor, EnableFloating, and all the PIL bits.
 	 * Also puts us in register window zero with traps off.
 	 */
+#if 0
 	set	(PSR_PS | PSR_S | PSR_PIL | PSR_EF), %g2
 	wr	%g2, 0x0, %psr
+#endif
 	
 	/* Load all registers
 	 */
 	set	__context, %g1
-        lda     [%g1] ASI_BP, %g1
-        add     %g1, 8, %g1     /* skip %g0 and %g1*/
+        ld      [%g1], %g1
+        ld      [%g1 + 8], %g2
+        ld      [%g1 + 12], %g3
+        ld      [%g1 + 16], %g4
+        ld      [%g1 + 20], %g5
+        ld      [%g1 + 24], %g6
+        ld      [%g1 + 28], %g7
 
-#define load_reg(reg) lda [%g1] ASI_BP, reg ; add %g1, 4, %g1
+        ld      [%g1 + 32], %o0
+        ld      [%g1 + 36], %o1
+        ld      [%g1 + 40], %o2
+        ld      [%g1 + 44], %o3
+        ld      [%g1 + 48], %o4
+        ld      [%g1 + 52], %o5
+        ld      [%g1 + 56], %o6
+        ld      [%g1 + 60], %o7
 
-        load_reg(%g2)
-        load_reg(%g3)
-        load_reg(%g4)
-        load_reg(%g5)
-        load_reg(%g6)
-        load_reg(%g7)
+        ld      [%g1 + 64], %l0
+        ld      [%g1 + 68], %l1
+        ld      [%g1 + 72], %l2
+        ld      [%g1 + 76], %l3
+        ld      [%g1 + 80], %l4
+        ld      [%g1 + 84], %l5
+        ld      [%g1 + 88], %l6
+        ld      [%g1 + 92], %l7
 
-        load_reg(%o0)
-        load_reg(%o1)
-        load_reg(%o2)
-        load_reg(%o3)
-        load_reg(%o4)
-        load_reg(%o5)
-        load_reg(%o6)
-        load_reg(%o7)
-
-        load_reg(%l0)
-        load_reg(%l1)
-        load_reg(%l2)
-        load_reg(%l3)
-        load_reg(%l4)
-        load_reg(%l5)
-        load_reg(%l6)
-        load_reg(%l7)
-
-        load_reg(%i0)
-        load_reg(%i1)
-        load_reg(%i2)
-        load_reg(%i3)
-        load_reg(%i4)
-        load_reg(%i5)
-        load_reg(%i6)
-        load_reg(%i7)
-
+        ld      [%g1 + 96], %i0
+        ld      [%g1 + 100], %i1
+        ld      [%g1 + 104], %i2
+        ld      [%g1 + 108], %i3
+        ld      [%g1 + 112], %i4
+        ld      [%g1 + 116], %i5
+        ld      [%g1 + 120], %i6
+        ld      [%g1 + 124], %i7
+        
+        ld      [%g1 + 128], %g1
        	/* Finally, load new %pc */
-        lda     [%g1] ASI_BP, %g1
         jmp     %g1
-	 nop
+         clr    %g1
 
 __exit_context:
 	/* Get back to the original context */
Index: arch/sparc32/context.c
===================================================================
--- arch/sparc32/context.c	(revision 23)
+++ arch/sparc32/context.c	(working copy)
@@ -20,8 +20,8 @@
  * It is placed at the bottom of our stack, and loaded by assembly routine
  * to start us up.
  */
-struct context main_ctx __attribute__((section (".initctx"))) = {
-    .regs[REG_SP] = (uint32_t) SP_LOC(&main_ctx),
+struct context main_ctx = {
+    .regs[REG_SP] = (uint32_t) &_estack - 96,
     .pc = (uint32_t) start_main,
     .npc = (uint32_t) start_main + 4,
     .return_addr = (uint32_t) __exit_context,
Index: drivers/pgtsrmmu.h
===================================================================
--- drivers/pgtsrmmu.h	(revision 23)
+++ drivers/pgtsrmmu.h	(working copy)
@@ -83,7 +83,7 @@
 #ifndef __ASSEMBLY__
 
 /* Accessing the MMU control register. */
-extern __inline__ unsigned int srmmu_get_mmureg(void)
+static __inline__ unsigned int srmmu_get_mmureg(void)
 {
         unsigned int retval;
 	__asm__ __volatile__("lda [%%g0] %1, %0\n\t" :
@@ -92,14 +92,14 @@
 	return retval;
 }
 
-extern __inline__ void srmmu_set_mmureg(unsigned long regval)
+static __inline__ void srmmu_set_mmureg(unsigned long regval)
 {
 	__asm__ __volatile__("sta %0, [%%g0] %1\n\t" : :
 			     "r" (regval), "i" (ASI_M_MMUREGS) : "memory");
 
 }
 
-extern __inline__ void srmmu_set_ctable_ptr(unsigned long paddr)
+static __inline__ void srmmu_set_ctable_ptr(unsigned long paddr)
 {
 	paddr = ((paddr >> 4) & SRMMU_CTX_PMASK);
 	__asm__ __volatile__("sta %0, [%1] %2\n\t" : :
@@ -108,7 +108,7 @@
 			     "memory");
 }
 
-extern __inline__ unsigned long srmmu_get_ctable_ptr(void)
+static __inline__ unsigned long srmmu_get_ctable_ptr(void)
 {
 	unsigned int retval;
 
@@ -119,14 +119,14 @@
 	return (retval & SRMMU_CTX_PMASK) << 4;
 }
 
-extern __inline__ void srmmu_set_context(int context)
+static __inline__ void srmmu_set_context(int context)
 {
 	__asm__ __volatile__("sta %0, [%1] %2\n\t" : :
 			     "r" (context), "r" (SRMMU_CTX_REG),
 			     "i" (ASI_M_MMUREGS) : "memory");
 }
 
-extern __inline__ int srmmu_get_context(void)
+static __inline__ int srmmu_get_context(void)
 {
 	register int retval;
 	__asm__ __volatile__("lda [%1] %2, %0\n\t" :
@@ -136,7 +136,7 @@
 	return retval;
 }
 
-extern __inline__ unsigned int srmmu_get_fstatus(void)
+static __inline__ unsigned int srmmu_get_fstatus(void)
 {
 	unsigned int retval;
 
@@ -146,7 +146,7 @@
 	return retval;
 }
 
-extern __inline__ unsigned int srmmu_get_faddr(void)
+static __inline__ unsigned int srmmu_get_faddr(void)
 {
 	unsigned int retval;
 
@@ -157,7 +157,7 @@
 }
 
 /* This is guaranteed on all SRMMU's. */
-extern __inline__ void srmmu_flush_whole_tlb(void)
+static __inline__ void srmmu_flush_whole_tlb(void)
 {
 	__asm__ __volatile__("sta %%g0, [%0] %1\n\t": :
 			     "r" (0x400),        /* Flush entire TLB!! */
@@ -166,7 +166,7 @@
 }
 
 /* These flush types are not available on all chips... */
-extern __inline__ void srmmu_flush_tlb_ctx(void)
+static __inline__ void srmmu_flush_tlb_ctx(void)
 {
 	__asm__ __volatile__("sta %%g0, [%0] %1\n\t": :
 			     "r" (0x300),        /* Flush TLB ctx.. */
@@ -174,7 +174,7 @@
 
 }
 
-extern __inline__ void srmmu_flush_tlb_region(unsigned long addr)
+static __inline__ void srmmu_flush_tlb_region(unsigned long addr)
 {
 	addr &= SRMMU_PGDIR_MASK;
 	__asm__ __volatile__("sta %%g0, [%0] %1\n\t": :
@@ -184,7 +184,7 @@
 }
 
 
-extern __inline__ void srmmu_flush_tlb_segment(unsigned long addr)
+static __inline__ void srmmu_flush_tlb_segment(unsigned long addr)
 {
 	addr &= SRMMU_PMD_MASK;
 	__asm__ __volatile__("sta %%g0, [%0] %1\n\t": :
@@ -193,7 +193,7 @@
 
 }
 
-extern __inline__ void srmmu_flush_tlb_page(unsigned long page)
+static __inline__ void srmmu_flush_tlb_page(unsigned long page)
 {
 	page &= PAGE_MASK;
 	__asm__ __volatile__("sta %%g0, [%0] %1\n\t": :
@@ -202,7 +202,7 @@
 
 }
 
-extern __inline__ unsigned long srmmu_hwprobe(unsigned long vaddr)
+static __inline__ unsigned long srmmu_hwprobe(unsigned long vaddr)
 {
 	unsigned long retval;
 
@@ -214,7 +214,7 @@
 	return retval;
 }
 
-extern __inline__ int
+static __inline__ int
 srmmu_get_pte (unsigned long addr)
 {
 	register unsigned long entry;
Index: drivers/iommu.c
===================================================================
--- drivers/iommu.c	(revision 23)
+++ drivers/iommu.c	(working copy)
@@ -18,7 +18,6 @@
 #include "pgtsrmmu.h"
 #include "iommu.h"
 
-#define PHYS_JJ_IOMMU 0x10000000 /* First page of sun4m IOMMU */
 #define IOPERM        (IOPTE_CACHE | IOPTE_WRITE | IOPTE_VALID)
 #define MKIOPTE(phys) (((((phys)>>4) & IOPTE_PAGE) | IOPERM) & ~IOPTE_WAZ)
 #define LOWMEMSZ 32 * 1024 * 1024
@@ -61,7 +60,8 @@
 
 #define NCTX_SWIFT  0x100
 
-static void iommu_init(struct iommu *t);
+static void iommu_init(struct iommu *t, unsigned long base);
+
 static void
 iommu_invalidate(struct iommu_regs *regs) {
     regs->tlbflush = 0;
@@ -70,7 +70,7 @@
 /*
  * Allocate memory. This is reusable.
  */
-static void
+void
 mem_init(struct mem *t, char *begin, char *limit)
 {
     t->start = begin;
@@ -82,8 +82,14 @@
 mem_alloc(struct mem *t, int size, int align)
 {
     char *p;
+    unsigned long pa;
 
-    p = (char *)((((unsigned int)t->curp) + (align-1)) & ~(align-1));
+    // The alignment restrictions refer to physical, not virtual
+    // addresses
+    pa = va2pa((unsigned long)t->curp) + (align - 1);
+    pa &= ~(align - 1);
+    p = (char *)pa2va(pa);
+    
     if (p >= t->uplim || p + size > t->uplim)
         return 0;
     t->curp = p + size;
@@ -193,12 +199,72 @@
  * Switch page tables.
  */
 void
-init_mmu_swift()
+init_mmu_swift(unsigned long base)
 {
+    extern unsigned int qemu_mem_size;
     unsigned int addr, i;
     unsigned long pa, va;
 
-    mem_init(&cmem, (char *) &_vmem, (char *)&_evmem);
+    push_str("/memory");
+    fword("find-device");
+
+    PUSH(0);
+    fword("encode-int");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(qemu_mem_size);
+    fword("encode-int");
+    fword("encode+");
+    push_str("reg");
+    fword("property");
+
+    PUSH(0);
+    fword("encode-int");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(va2pa((unsigned long)&_data) - 1);
+    fword("encode-int");
+    fword("encode+");
+    push_str("available");
+    fword("property");
+
+    push_str("/virtual-memory");
+    fword("find-device");
+
+    PUSH(0);
+    fword("encode-int");
+    PUSH(base);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(0x300);
+    fword("encode-int");
+    fword("encode+");
+    push_str("reg");
+    fword("property");
+
+    PUSH(0);
+    fword("encode-int");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(va2pa((unsigned long)&_start) - 1);
+    fword("encode-int");
+    fword("encode+");
+
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(va2pa((unsigned long)&_iomem));
+    fword("encode-int");
+    fword("encode+");
+    PUSH(-va2pa((unsigned long)&_iomem));
+    fword("encode-int");
+    fword("encode+");
+    push_str("available");
+    fword("property");
+
     mem_init(&cio, (char *)&_end, (char *)&_iomem);
 
     context_table = mem_zalloc(&cmem, NCTX_SWIFT * sizeof(int), NCTX_SWIFT * sizeof(int));
@@ -244,7 +310,7 @@
     srmmu_set_context(0);
     srmmu_set_ctable_ptr(va2pa((unsigned long)context_table));
     srmmu_flush_whole_tlb();
-    iommu_init(&ciommu);
+    iommu_init(&ciommu, base);
 }
 /*
  * XXX This is a problematic interface. We alloc _memory_ which is uncached.
@@ -306,7 +372,7 @@
  * the routine is higher in food chain.
  */
 static void
-iommu_init(struct iommu *t)
+iommu_init(struct iommu *t, unsigned long base)
 {
     unsigned int *ptab;
     int ptsize;
@@ -314,7 +380,7 @@
     unsigned int impl, vers;
     unsigned int tmp;
 
-    if ((regs = map_io(PHYS_JJ_IOMMU, sizeof(struct iommu_regs))) == 0) {
+    if ((regs = map_io(base, sizeof(struct iommu_regs))) == 0) {
         DPRINTF("Cannot map IOMMU\n");
         for (;;) { }
     }
@@ -343,11 +409,12 @@
 
     /* flush_cache_all(); */
     /** flush_tlb_all(); **/
-    regs->base = ((unsigned int)va2pa((unsigned long)ptab)) >> 4;
+    tmp = (unsigned int)va2pa((unsigned long)ptab);
+    regs->base = tmp >> 4;
     iommu_invalidate(regs);
 
-    DPRINTF("IOMMU: impl %d vers %d page table at 0x%p of size %d bytes\n",
-            impl, vers, t->page_table, ptsize);
+    DPRINTF("IOMMU: impl %d vers %d page table at 0x%p (pa 0x%x) of size %d bytes\n",
+            impl, vers, t->page_table, tmp, ptsize);
 
     mem_init(&t->bmap, (char*)t->plow, (char *)0xfffff000);
 }
Index: drivers/esp.c
===================================================================
--- drivers/esp.c	(revision 23)
+++ drivers/esp.c	(working copy)
@@ -25,8 +25,9 @@
 #include "asm/dma.h"
 #include "esp.h"
 
-#define PHYS_JJ_ESPDMA  0x78400000      /* ESP DMA controller */
-#define PHYS_JJ_ESP     0x78800000      /* ESP SCSI */
+#define MACIO_ESPDMA  0x08400000      /* ESP DMA controller */
+#define MACIO_ESP     0x08800000      /* ESP SCSI */
+
 #define BUFSIZE         4096
 
 #define REGISTER_NAMED_NODE( name, path )   do {                        \
@@ -263,7 +264,10 @@
     id = POP();
     *sd = &global_esp->sd[id];
 
-    DPRINTF("opening drive %d\n", id);
+#ifdef CONFIG_DEBUG_ESP
+    fword("my-args");
+    DPRINTF("opening drive %d args %s\n", id, pop_fstr_copy());
+#endif
 
     selfword("open-deblocker");
 
@@ -291,16 +295,14 @@
 
 
 static int
-espdma_init(struct esp_dma *espdma)
+espdma_init(unsigned long base, struct esp_dma *espdma)
 {
-    void *p;
+    espdma->regs = (void *)map_io(base + MACIO_ESPDMA, 0x10);
 
-    /* Hardcode everything for MrCoffee. */
-    if ((p = (void *)map_io(PHYS_JJ_ESPDMA, 0x10)) == 0) {
+    if (espdma->regs == 0) {
         DPRINTF("espdma_init: cannot map registers\n");
         return -1;
     }
-    espdma->regs = p;
 
     DPRINTF("dma1: ");
 
@@ -337,6 +339,21 @@
     }
     DPRINTF("\n");
 
+    push_str("/iommu/sbus/espdma");
+    fword("find-device");
+
+    /* set reg */
+    PUSH(4);
+    fword("encode-int");
+    PUSH(MACIO_ESPDMA);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(0x00000010);
+    fword("encode-int");
+    fword("encode+");
+    push_str("reg");
+    fword("property");
+
     return 0;
 }
 
@@ -355,7 +372,7 @@
     /* set reg */
     PUSH(4);
     fword("encode-int");
-    PUSH(0x08800000);
+    PUSH(MACIO_ESP);
     fword("encode-int");
     fword("encode+");
     PUSH(0x00000010);
@@ -393,7 +410,7 @@
 };
 
 static void
-add_alias(const unsigned char *device, const unsigned char *alias)
+add_alias(const char *device, const char *alias)
 {
     push_str("/aliases");
     fword("find-device");
@@ -404,7 +421,8 @@
     fword("property");
 }
 
-int ob_esp_init(void)
+int
+ob_esp_init(unsigned long base)
 {
     int id, diskcount = 0, cdcount = 0, *counter_ptr;
     char nodebuff[256], aliasbuff[256];
@@ -413,13 +431,18 @@
     DPRINTF("Initializing SCSI...");
 
     esp = malloc(sizeof(esp_private_t));
+    if (!esp) {
+        DPRINTF("Can't allocate ESP private structure\n");
+        return -1;
+    }
+
     global_esp = esp;
 
-    if (espdma_init(&esp->espdma) != 0) {
+    if (espdma_init(base, &esp->espdma) != 0) {
         return -1;
     }
     /* Get the IO region */
-    esp->ll = (void *)map_io(PHYS_JJ_ESP, sizeof(struct esp_regs));
+    esp->ll = (void *)map_io(base + MACIO_ESP, sizeof(struct esp_regs));
     if (esp->ll == 0) {
         DPRINTF("Can't map ESP registers\n");
         return -1;
@@ -433,7 +456,9 @@
 
     // Chip reset
     esp->ll->regs[ESP_CMD] = ESP_CMD_RC;
-
+    
+    DPRINTF("ESP at 0x%lx, buffer va 0x%lx dva 0x%lx\n", (unsigned long)esp,
+            (unsigned long)esp->buffer, (unsigned long)esp->buffer_dvma);
     DPRINTF("done\n");
     DPRINTF("Initializing SCSI devices...");
 
@@ -448,8 +473,7 @@
 #endif
     }
 
-    sprintf(nodebuff, "/iommu/sbus/espdma/esp");
-    REGISTER_NAMED_NODE(ob_esp, nodebuff);
+    REGISTER_NAMED_NODE(ob_esp, "/iommu/sbus/espdma/esp");
     device_end();
 
     for (id = 0; id < 8; id++) {
Index: drivers/obio.c
===================================================================
--- drivers/obio.c	(revision 23)
+++ drivers/obio.c	(working copy)
@@ -17,69 +17,395 @@
 #include "libc/vsprintf.h"
 
 #include "openbios/drivers.h"
+#include "openbios/nvram.h"
+#include "obio.h"
 
+#define REGISTER_NAMED_NODE( name, path )   do { \
+	     bind_new_node( name##_flags_, name##_size_, \
+			path, name##_m, sizeof(name##_m)/sizeof(method_t)); \
+	} while(0)
+
+
+/* DECLARE data structures for the nodes.  */
+DECLARE_UNNAMED_NODE( ob_obio, INSTALL_OPEN, sizeof(int) );
+
 static void
-ob_zs_init(unsigned long slavio_base)
+ob_new_obio_device(const char *name, const char *type)
 {
-    unsigned long zs_addr;
+    push_str("/obio");
+    fword("find-device");
+    fword("new-device");
 
-    zs_addr = (unsigned long)map_io(slavio_base + 0x0, 8);
-    push_str("/obio/zs@0,0");
-    fword("find-device");
-    
-    PUSH(zs_addr);
+    push_str(name);
+    fword("device-name");
+
+    if (type) {
+        push_str(type);
+        fword("device-type");
+    }
+}
+
+static unsigned long
+ob_reg(unsigned long base, unsigned long offset, unsigned long size, int map)
+{
+    PUSH(0);
     fword("encode-int");
-    PUSH(sizeof(zs_addr));
+    PUSH(offset);
     fword("encode-int");
     fword("encode+");
-    push_str("address");
+    PUSH(size);
+    fword("encode-int");
+    fword("encode+");
+    push_str("reg");
     fword("property");
 
-    zs_addr = (unsigned long)map_io(slavio_base + 0x00100000, 8);
-    push_str("/obio/zs@0,100000");
+    if (map) {
+        unsigned long addr;
+
+        addr = (unsigned long)map_io(base + offset, size);
+
+        PUSH(addr);
+        fword("encode-int");
+        PUSH(4);
+        fword("encode-int");
+        fword("encode+");
+        push_str("address");
+        fword("property");
+        return addr;
+    }
+    return 0;
+}
+
+static void
+ob_intr(int intr)
+{
+    PUSH(intr);
+    fword("encode-int");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    push_str("intr");
+    fword("property");
+}
+
+
+static void
+ob_zs_init(unsigned long base, unsigned long offset, int intr, int slave, int keyboard)
+{
+    ob_new_obio_device("zs", "serial");
+
+    ob_reg(base, offset, ZS_REGS, 1);
+
+    PUSH(slave);
+    fword("encode-int");
+    push_str("slave");
+    fword("property");
+
+    if (keyboard) {
+        PUSH(-1);
+        fword("encode-int");
+        push_str("keyboard");
+        fword("property");
+
+        PUSH(-1);
+        fword("encode-int");
+        push_str("mouse");
+        fword("property");
+    }
+
+    ob_intr(intr);
+
+    fword("finish-device");
+}
+
+static char *nvram;
+struct qemu_nvram_v1 nv_info;
+
+void
+arch_nvram_get(char *data)
+{
+    memcpy(data, nvram, NVRAM_SIZE);
+}
+
+void
+arch_nvram_put(char *data)
+{
+    memcpy(nvram, data, NVRAM_SIZE);
+}
+
+int
+arch_nvram_size(void)
+{
+    return NVRAM_SIZE;
+}
+
+static void
+ob_nvram_init(unsigned long base, unsigned long offset)
+{
+    extern uint32_t kernel_image;
+    extern uint32_t kernel_size;
+    extern uint32_t cmdline;
+    extern uint32_t cmdline_size;
+    extern char boot_device;
+
+    ob_new_obio_device("eeprom", NULL);
+
+    nvram = (char *)ob_reg(base, offset, NVRAM_SIZE, 1);
+
+    memcpy(&nv_info, nvram, sizeof(nv_info));
+
+    printk("Nvram id %s, version %d\n", nv_info.id_string, nv_info.version);
+    if (strcmp(nv_info.id_string, "QEMU_BIOS") || nv_info.version != 1) {
+        printk("Unknown nvram, freezing!\n");
+        for (;;);
+    }
+    kernel_image = nv_info.kernel_image;
+    kernel_size = nv_info.kernel_size;
+    cmdline = nv_info.cmdline;
+    cmdline_size = nv_info.cmdline_size;
+    boot_device = nv_info.boot_device;
+
+    push_str("mk48t08");
+    fword("model");
+
+    fword("finish-device");
+
+    push_str("/");
     fword("find-device");
     
-    PUSH(zs_addr);
+    PUSH((long)&nvram[NVRAM_IDPROM]);
+    PUSH(32);
+    fword("encode-bytes");
+    push_str("idprom");
+    fword("property");
+}
+
+static void
+ob_fd_init(unsigned long base, unsigned long offset, int intr)
+{
+    ob_new_obio_device("SUNW,fdtwo", "block");
+
+    ob_reg(base, offset, FD_REGS, 0);
+
+    ob_intr(intr);
+
+    fword("finish-device");
+}
+
+
+static void
+ob_sconfig_init(unsigned long base, unsigned long offset)
+{
+    ob_new_obio_device("slavioconfig", NULL);
+
+    ob_reg(base, offset, SCONFIG_REGS, 0);
+
+    fword("finish-device");
+}
+
+static void
+ob_auxio_init(unsigned long base, unsigned long offset)
+{
+    ob_new_obio_device("auxio", NULL);
+
+    ob_reg(base, offset, AUXIO_REGS, 0);
+
+    fword("finish-device");
+}
+
+static void
+ob_power_init(unsigned long base, unsigned long offset, int intr)
+{
+    ob_new_obio_device("power", NULL);
+
+    ob_reg(base, offset, POWER_REGS, 0);
+
+    ob_intr(intr);
+
+    fword("finish-device");
+}
+
+static void
+ob_counter_init(unsigned long base, unsigned long offset)
+{
+    volatile struct sun4m_timer_regs *regs;
+
+    ob_new_obio_device("counter", NULL);
+
+    PUSH(0);
     fword("encode-int");
-    PUSH(sizeof(zs_addr));
+    PUSH(offset);
     fword("encode-int");
     fword("encode+");
-    push_str("address");
+    PUSH(COUNTER_REGS);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(offset + 0x10000);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(COUNTER_REGS);
+    fword("encode-int");
+    fword("encode+");
+    push_str("reg");
     fword("property");
+
+    fword("finish-device");
+
+    regs = map_io(base + offset, sizeof(*regs));
+    regs->l10_timer_limit = (((1000000/100) + 1) << 10);
+    regs->cpu_timers[0].l14_timer_limit = 0;
 }
 
 static void
-ob_obio_open(int *idx)
+ob_interrupt_init(unsigned long base, unsigned long offset)
 {
+    volatile struct sun4m_intregs *regs;
+
+    ob_new_obio_device("interrupt", NULL);
+
+    PUSH(0);
+    fword("encode-int");
+    PUSH(offset);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(INTERRUPT_REGS);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(offset + 0x10000);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(INTERRUPT_REGS);
+    fword("encode-int");
+    fword("encode+");
+    push_str("reg");
+    fword("property");
+
+    fword("finish-device");
+
+    regs = map_io(base + offset, sizeof(*regs));
+    regs->set = ~SUN4M_INT_MASKALL;
+    regs->cpu_intregs[0].clear = ~0x17fff;
+}
+
+
+static void
+ob_obio_open(__attribute__((unused))int *idx)
+{
 	int ret=1;
 	RET ( -ret );
 }
 
 static void
-ob_obio_close(int *idx)
+ob_obio_close(__attribute__((unused))int *idx)
 {
 	selfword("close-deblocker");
 }
 
 static void
-ob_obio_initialize(int *idx)
+ob_obio_initialize(__attribute__((unused))int *idx)
 {
+    push_str("/");
+    fword("find-device");
+    fword("new-device");
+
+    push_str("obio");
+    fword("device-name");
+
+    push_str("hierarchical");
+    fword("device-type");
+
+    PUSH(2);
+    fword("encode-int");
+    push_str("#address-cells");
+    fword("property");
+
+    PUSH(1);
+    fword("encode-int");
+    push_str("#size-cells");
+    fword("property");
+
+    fword("finish-device");
 }
 
+static void
+ob_set_obio_ranges(unsigned long base)
+{
+    push_str("/obio");
+    fword("find-device");
+    PUSH(0);
+    fword("encode-int");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(0);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(base);
+    fword("encode-int");
+    fword("encode+");
+    PUSH(SLAVIO_SIZE);
+    fword("encode-int");
+    fword("encode+");
+    push_str("ranges");
+    fword("property");
+}
 
-NODE_METHODS(ob_obio_node) = {
+static void
+ob_obio_decodeunit(__attribute__((unused)) int *idx)
+{
+    fword("decode-unit-sbus");
+}
+
+
+static void
+ob_obio_encodeunit(__attribute__((unused)) int *idx)
+{
+    fword("encode-unit-sbus");
+}
+
+NODE_METHODS(ob_obio) = {
 	{ NULL,			ob_obio_initialize	},
 	{ "open",		ob_obio_open		},
 	{ "close",		ob_obio_close		},
+	{ "encode-unit",	ob_obio_encodeunit	},
+	{ "decode-unit",	ob_obio_decodeunit	},
 };
 
 
-int ob_obio_init(unsigned long slavio_base)
+int
+ob_obio_init(unsigned long slavio_base)
 {
 
     //printk("Initializing OBIO devices...\n");
-    
-    ob_zs_init(slavio_base);
-	
+#if 0 // XXX
+    REGISTER_NAMED_NODE(ob_obio, "/obio");
+    device_end();
+    ob_set_obio_ranges(slavio_base);
+#endif
+
+    ob_zs_init(slavio_base, SLAVIO_ZS, ZS_INTR, 1, 1);
+
+    ob_zs_init(slavio_base, SLAVIO_ZS1, ZS_INTR, 0, 0);
+
+    ob_nvram_init(slavio_base, SLAVIO_NVRAM);
+
+    ob_fd_init(slavio_base, SLAVIO_FD, FD_INTR);
+
+    ob_sconfig_init(slavio_base, SLAVIO_SCONFIG);
+
+    ob_auxio_init(slavio_base, SLAVIO_AUXIO);
+
+    ob_power_init(slavio_base, SLAVIO_POWER, POWER_INTR);
+
+    ob_counter_init(slavio_base, SLAVIO_COUNTER);
+
+    ob_interrupt_init(slavio_base, SLAVIO_INTERRUPT);
+
     return 0;
 }
Index: drivers/obio.h
===================================================================
--- drivers/obio.h	(revision 0)
+++ drivers/obio.h	(revision 0)
@@ -0,0 +1,187 @@
+/* Addresses, interrupt numbers, register sizes */
+
+#define SLAVIO_ZS        0x00000000
+#define SLAVIO_ZS1       0x00100000
+#define ZS_INTR          0x2c
+#define ZS_REGS          8
+
+#define SLAVIO_NVRAM     0x00200000
+#define NVRAM_SIZE       0x2000
+#define NVRAM_IDPROM     0x1fd8
+
+#define SLAVIO_FD        0x00400000
+#define FD_REGS          15
+#define FD_INTR          0x2b
+
+#define SLAVIO_SCONFIG   0x00800000
+#define SCONFIG_REGS     1
+
+#define SLAVIO_AUXIO     0x00900000
+#define AUXIO_REGS       1
+
+#define SLAVIO_POWER     0x00910000
+#define POWER_REGS       1
+#define POWER_INTR       0x22
+
+#define SLAVIO_COUNTER   0x00d00000
+#define COUNTER_REGS     0x10
+
+#define SLAVIO_INTERRUPT 0x00e00000
+#define INTERRUPT_REGS   0x10
+
+#define SLAVIO_SIZE      0x01000000
+
+struct qemu_nvram_v1 {
+    char id_string[16];
+    uint32_t version;
+    uint32_t nvram_size; // not used in Sun4m
+    char unused1[8];
+    char arch[12];
+    char smp_cpus;
+    char unused2;
+    char nographic;
+    uint32_t ram_size;
+    char boot_device;
+    char unused3[3];
+    uint32_t kernel_image;
+    uint32_t kernel_size;
+    uint32_t cmdline;
+    uint32_t cmdline_size;
+    uint32_t initrd_image;
+    uint32_t initrd_size;
+    uint32_t nvram_image;
+    uint16_t width;
+    uint16_t height;
+    uint16_t depth;
+    char unused4[158];
+    uint16_t crc;
+};
+
+#define SUN4M_NCPUS      16
+#define PAGE_SIZE        4096
+
+/* linux/include/asm-sparc/timer.h */
+
+/* A sun4m has two blocks of registers which are probably of the same
+ * structure. LSI Logic's L64851 is told to _decrement_ from the limit
+ * value. Aurora behaves similarly but its limit value is compacted in
+ * other fashion (it's wider). Documented fields are defined here.
+ */
+
+/* As with the interrupt register, we have two classes of timer registers
+ * which are per-cpu and master.  Per-cpu timers only hit that cpu and are
+ * only level 14 ticks, master timer hits all cpus and is level 10.
+ */
+
+#define SUN4M_PRM_CNT_L       0x80000000
+#define SUN4M_PRM_CNT_LVALUE  0x7FFFFC00
+
+struct sun4m_timer_percpu_info {
+  __volatile__ unsigned int l14_timer_limit;    /* Initial value is 0x009c4000 */
+  __volatile__ unsigned int l14_cur_count;
+
+  /* This register appears to be write only and/or inaccessible
+   * on Uni-Processor sun4m machines.
+   */
+  __volatile__ unsigned int l14_limit_noclear;  /* Data access error is here */
+
+  __volatile__ unsigned int cntrl;            /* =1 after POST on Aurora */
+  __volatile__ unsigned char space[PAGE_SIZE - 16];
+};
+
+struct sun4m_timer_regs {
+	struct sun4m_timer_percpu_info cpu_timers[SUN4M_NCPUS];
+	volatile unsigned int l10_timer_limit;
+	volatile unsigned int l10_cur_count;
+
+	/* Again, this appears to be write only and/or inaccessible
+	 * on uni-processor sun4m machines.
+	 */
+	volatile unsigned int l10_limit_noclear;
+
+	/* This register too, it must be magic. */
+	volatile unsigned int foobar;
+
+	volatile unsigned int cfg;     /* equals zero at boot time... */
+};
+
+/*
+ * Registers of hardware timer in sun4m.
+ */
+struct sun4m_timer_percpu {
+    volatile unsigned int l14_timer_limit; /* Initial value is 0x009c4000 */
+    volatile unsigned int l14_cur_count;
+};
+
+struct sun4m_timer_global {
+    volatile unsigned int l10_timer_limit;
+    volatile unsigned int l10_cur_count;
+};
+
+/* linux/include/asm-sparc/irq.h */
+
+/* These registers are used for sending/receiving irqs from/to
+ * different cpu's.
+ */
+struct sun4m_intreg_percpu {
+    unsigned int tbt;        /* Interrupts still pending for this cpu. */
+
+    /* These next two registers are WRITE-ONLY and are only
+     * "on bit" sensitive, "off bits" written have NO affect.
+     */
+    unsigned int clear;  /* Clear this cpus irqs here. */
+    unsigned int set;    /* Set this cpus irqs here. */
+    unsigned char space[PAGE_SIZE - 12];
+};
+
+/*
+ * djhr
+ * Actually the clear and set fields in this struct are misleading..
+ * according to the SLAVIO manual (and the same applies for the SEC)
+ * the clear field clears bits in the mask which will ENABLE that IRQ
+ * the set field sets bits in the mask to DISABLE the IRQ.
+ *
+ * Also the undirected_xx address in the SLAVIO is defined as
+ * RESERVED and write only..
+ *
+ * DAVEM_NOTE: The SLAVIO only specifies behavior on uniprocessor
+ *             sun4m machines, for MP the layout makes more sense.
+ */
+struct sun4m_intregs {
+    struct sun4m_intreg_percpu cpu_intregs[SUN4M_NCPUS];
+    unsigned int tbt;                /* IRQ's that are still pending. */
+    unsigned int irqs;               /* Master IRQ bits. */
+
+    /* Again, like the above, two these registers are WRITE-ONLY. */
+    unsigned int clear;              /* Clear master IRQ's by setting bits here. */
+    unsigned int set;                /* Set master IRQ's by setting bits here. */
+
+    /* This register is both READ and WRITE. */
+    unsigned int undirected_target;  /* Which cpu gets undirected irqs. */
+};
+
+/* Dave Redman (djhr@tadpole.co.uk)
+ * The sun4m interrupt registers.
+ */
+#define SUN4M_INT_ENABLE  	0x80000000
+#define SUN4M_INT_E14     	0x00000080
+#define SUN4M_INT_E10     	0x00080000
+
+#define SUN4M_HARD_INT(x)	(0x000000001 << (x))
+#define SUN4M_SOFT_INT(x)	(0x000010000 << (x))
+
+#define	SUN4M_INT_MASKALL	0x80000000	  /* mask all interrupts */
+#define	SUN4M_INT_MODULE_ERR	0x40000000	  /* module error */
+#define	SUN4M_INT_M2S_WRITE	0x20000000	  /* write buffer error */
+#define	SUN4M_INT_ECC		0x10000000	  /* ecc memory error */
+#define	SUN4M_INT_FLOPPY	0x00400000	  /* floppy disk */
+#define	SUN4M_INT_MODULE	0x00200000	  /* module interrupt */
+#define	SUN4M_INT_VIDEO		0x00100000	  /* onboard video */
+#define	SUN4M_INT_REALTIME	0x00080000	  /* system timer */
+#define	SUN4M_INT_SCSI		0x00040000	  /* onboard scsi */
+#define	SUN4M_INT_AUDIO		0x00020000	  /* audio/isdn */
+#define	SUN4M_INT_ETHERNET	0x00010000	  /* onboard ethernet */
+#define	SUN4M_INT_SERIAL	0x00008000	  /* serial ports */
+#define	SUN4M_INT_KBDMS		0x00004000	  /* keyboard/mouse */
+#define	SUN4M_INT_SBUSBITS	0x00003F80	  /* sbus int bits */
+
